con
  _clkmode = xtal1 + pll16x      'set to ext low speed crystal
  _xinfreq = 5_000_000          'Frequency on XIN pin is 5 MHz

  Start_byte = 0               'serial communication start byte
  
var                             
  'these variables are unique to each instance of this object.

  long Num_ch                   'number of servos (physical_channels - 1)
  long serial_rx_lock           'lock number of the application port RX lock
  long serial_tx_lock           'lock number of the application port TX lock
  'long tx_head_p                'pointer to the tx_head counter of the serial object
  'long tx_buffer_p              'pointer to the start of the tx_buffer in the serial object
  long LPOK_thresh              'contains the noninclusive upper limit for the P to P delay [in clocks]
  long mute                     'mute flag that stops the LPOK detector from transmitting if TRUE
  long m_wait                   'variable used to delay the setting of mute
  long bias                     'value subtracted from pulse to pulse delay measurements and added to LPOK_thresh 
  long stack1[150]              'some stack space
  long stack2[150]              'once features are finalised, investigate how much stack is really needed
  long stack3[150]
  long stack4[150]
  long stack5[150]

obj
  serial : "fullduplexserial_ASM_mod.spin"                      'serial port
  servo : "servo_diagnostics_2"                         'make servo pulses and use servo position detection circuit
  lpok : "LPOK detector S.spin"                         'times Q-switch to laser pulse time delay and generates LPOK logic signal
  {this top object will start several copies of the commander code in new cogs, and share a single serial object}                     

Pub main
  {{it all starts here}}
  'setup serial port
  serial.start(27, 26, 0, 115200) 'start my application port
  serial_rx_lock := locknew     'get a lock for sharing the RX buffer
  serial_tx_lock := locknew     'get a lock for sharing the TX buffer
  serial.str(string("Good morning"))                    'show some signs of life
  serial.tx($0D)

  'setup servo controlling object
  Num_ch := servo.Num_ch_con    'pull in the number of active channels 
  servo.init                    'setup the timing variables for the servos

  'setup LPOK code
  mute := true                  'LPOK detector is muted by default
  m_wait := mute                'preload delay variable
  LPOK_thresh := 200            'start out with a high threshold.
  bias := 0<<7                  'start with an initial bias value of zero                             
  lpok.start(24,23,25,@LPOK_thresh, @mute, serial_tx_lock, serial.tx_head_p, serial.tx_buffer_p,@bias)         'start LPOK detector cog.
  {8,10,14 for testing}{*&^% pins used for testing are different from final board!*&^%}

  {{start loading up the "commander" cogs}} 
  cognew(commander("A"),@stack1)     'fully locked and arbitrated now
  cognew(commander("B"),@stack2)
  cognew(commander("C"),@stack3)
  cognew(commander("D"),@stack4)
  'cognew(commander("E"),@stack5)
  commander("F")

  


  
Pub commander(tag) | temp, data[3], i, status, dump, temp2
{{this is the main command processor.  it watches the serial port for a start byte.
Grabs a packet.  Verifies the packet.  then interperts the command and data.  afterwards it returns
a packet containing the finished command and the status_code of that completed command}}
{assumes Init has been run once and FullDuplexSerial is running}
{one start byte should be reserved as the "reset" byte.  just send more than three of this byte to
reset the state of all controllers}
  dump := 0                     'clear dump flag
  m_wait := true                'start off muted by default

  repeat
    temp := lockset(serial_rx_lock)    'wait for the application port rx buffer to be free
  while temp            
  repeat
    temp := serial.rx           'grab a byte
    if temp == Start_byte       'is it the start byte?
      repeat i from 0 to 2      'get the rest of the packet
        temp += data[i] := serial.rx
      temp -= data[2]          'remove x-mitted checksum from accumulator
      temp //= 256             'compute a new checksum
      if data[2] == temp                                'does it match the sent sum?
        lockclr(serial_rx_lock) 'i'm done with the application port's RX buffer, free it up
        status := 0             'clear the status variable
        case data[0]
          0..Num_ch:            'move commands
            status := servo.move_servo(data[0], data[1])        'move the servo and update the status
          16:                   'find current position
            status := servo.search_servo(data[1])       'search out current pos and return it.
          18..(Num_ch+18):      'force move command
            status := servo.force_move(data[0]-18, data[1])                     'force a servo move and return the final position
          34..(Num_ch+34):
            status := servo.set_low_limit(data[0]-34, data[1])                  'set the lower software movement limit
          50..(Num_ch+50):
            status := servo.set_high_limit(data[0]-50, data[1])                 'set the high software movement limit
          66..(Num_ch+66):
            status := servo.set_speed(data[0]-66, data[1])                      'set the speed of movement
          84:
            if data[1] == 172
              reboot                                                            'software controlled reset
            else
              status := 16                                                     'reset command bad
          82:
            servo.Save_servo_data                                               'write the updated servo info structure to EEPROM.
            status := 0
          86:                                                                   'dump the servo setup array
            dump := 1                                                           'set dump flag
            status := 0
          88:                                                                   'mute or un-mute the reporting of Q-switch to Laser-pulse delays
            m_wait := data[1]                                                   'load command payload into
            status := m_wait                                                    'echo command payload as status byte
          90:                                                                   'set the bias value added to LPOK_thresh and subtracted from raw data
            bias := data[1] << 7                                                'multiply by 2^7
            status := data[1]                                                   'echo command payload as status byte
          92:                                                                   'set the threashold for Q to APD delay below which a LPOK pulse is generated
            LPOK_thresh := 1 #> data[1] <# 255                                  'set the variable with limit checking
            status := data[1]                                                   'echo command payload as status byte.                                                                                                                                                        
          other:                'don't know what u sent crazy man!
            status += 16                                'bad command
        repeat
          temp := lockset(serial_tx_lock)  'wait for the application port's tx buffer to be free
        while temp              'then send a response packet
        if dump == 1
          dump := 0             'clear the flag 'cause we're doing it right now
          serial.str(string("Channel,Speed,Lo-limit,Hi-limit"))
          serial.tx($0D)        'carrage return
          repeat i from 0 to Num_ch
            temp := servo.servo_dat_point(i)        'pointer to the start of a line in the setup array
            serial.dec(i)
            serial.tx($2C)      'comma delimiter
            temp2 := long[temp][1]
            serial.dec(temp2)   'print the speed                        
            serial.tx($2C)      'comma delimiter
            temp2 := long[temp][2]
            serial.dec(temp2)   'print the lo-limit                        
            serial.tx($2C)      'comma delimiter
            temp2 := long[temp][3]
            serial.dec(temp2)   'print the Hi-limit
            serial.tx($0D)      'carrage return
                      
        serial.tx(Start_byte)   'start byte
        serial.tx(data[0])      'echo command
        serial.tx(status)       'send status code
        temp := (Start_byte + data[0] + status) // 256
        serial.tx(temp)         'send a checksum        
        lockclr(serial_tx_lock)             'free up the tx buffer
        mute := m_wait          'update mute flag
        repeat
          temp := lockset(serial_rx_lock)  'wait for the application port's rx buffer to be free.  
        while temp
      else
        lockclr(serial_rx_lock) 'i'm done with the application port's RX buffer, free it up
        repeat
          temp := lockset(serial_tx_lock)  'wait for the application port's tx buffer to be free
        while temp              'then send a response packet
        serial.tx(Start_byte)   'start byte
        serial.tx(15)           'indicates a controller status packet
        serial.tx(32)           'bad packet
        serial.tx((Start_byte + 15 + 32)//256)          'check sum
        lockclr(serial_tx_lock)             'free up the tx buffer
        repeat
          temp := lockset(serial_rx_lock)  'wait for the application port's rx buffer to be free.  
        while temp
        